#import <Cocoa/Cocoa.h>
#import <Foundation/Foundation.h>
#import "mod_metaphor.h"
#import "mod_metaphor_debug.h"

/* metadata headers:  
* /System/Library/Frameworks/Foundation.framework/Versions/Current/Headers 
* /System/Library/Frameworks/CoreServices.framework/Versions/A/Frameworks/Metadata.framework/Versions/A/Headers 
*/


/*
@interface PropertyListCoder {
	NSMutableDictionary *data;
}

- (void)encodeValueOfObjCType:(const char *)valueType at:(const void *)address;
- (void)decodeValueOfObjCType:(const char *)valueType at:(void *)data;
@end

@implementation PropertyListCoder
- (id)init {
	if(self = [super init]) {
		data = [[NSMutableDictionary alloc] init];
	}
	
	return self;
}

- (void)encodeValueOfObjCType:(const char *)valueType at:(const void *)address {
	
}

- (void)decodeValueOfObjCType:(const char *)valueType at:(void *)data {
	
}
@end 
*/

@implementation NSMetadataItem(attributes)
- (NSDictionary *)attributeDictionary {
	NSMutableDictionary *item = [NSMutableDictionary dictionaryWithDictionary:[self valuesForAttributes:[self attributes]]];

	/* kMDItemPath isn't included in [metadataItem attributes] for some reason */
	NSString *path = [self valueForAttribute:@"kMDItemPath"];
	[item setObject:path forKey:@"kMDItemPath"];

	NSLogL1(@"Result Artist: %@, Album: %@, Path: %@",
																	[item objectForKey:@"kMDItemAuthors"], 
																	[item objectForKey:@"kMDItemAlbum"],
																	[item objectForKey:@"kMDItemPath"]);
																	
	return (NSDictionary *)item;
}
@end

@implementation NSArray(rawDescription)
- (NSString *)rawDescription {
	NSString *ret = [[[NSMutableString alloc] init] autorelease];
	NSEnumerator *enumerator = [self objectEnumerator];
	id item;
	
	while(item = [enumerator nextObject]) [ret appendString:[item description]];
	
	return ret;
}
@end

@implementation NSString(removeTrailingSlash)
- (NSString *)removeTrailingSlash {
	if([self hasSuffix:@"/"] == YES)
		return [self substringToIndex:([self length]-1)];
	return self;
}
@end

@implementation NSString(HTTPURLProcessing)
- (NSString *)stringByEncodingHTTPURL {
	return (NSString *)CFURLCreateStringByAddingPercentEscapes(kCFAllocatorDefault, (CFStringRef)self, NULL, (CFStringRef)@";/?:@&=+$,", kCFStringEncodingUTF8);
}

- (NSString *)stringByDecodingHTTPURL {
	return (NSString *)CFURLCreateStringByReplacingPercentEscapesUsingEncoding (kCFAllocatorDefault, (CFStringRef)self, (CFStringRef)@"", kCFStringEncodingUTF8);
}
@end

@implementation NSString(stringReplacement)
- (NSString *)stringWithStringReplaced:(NSString *)find with:(NSString *)replace {
	NSMutableString *ret = [NSMutableString stringWithString:self];
	[ret replaceOccurrencesOfString:find withString:replace options:nil range:NSMakeRange(0, [self length])];
	return (NSString *)ret;
}
@end

@implementation NameFormatter
- (void) encodeWithCoder:(NSCoder *)coder
{
	[coder encodeObject:string forKey:@"string"];
	[coder encodeInt:type forKey:@"type"];
}

- (id) initWithCoder:(NSCoder *)coder
{
	if (self = [self init])
	{
		[self setString:[coder decodeObjectForKey:@"string"]];
		[self setType:[coder decodeIntForKey:@"type"]];
	}

	return self;
}

- (id)initWithPropertyList:(NSDictionary *)plist {
	if(self = [self init]) {
		[self setString:[plist objectForKey:@"string"]];
		[self setType:[[plist objectForKey:@"type"] intValue]];
	}
	
	return self;
}

- (NSString *)description {
	switch([self type]) {
		case NameFormatterKey:
			return [NSString stringWithFormat:@"\@%@\@", [self string]];
		default:
			return [self string];
	}
}

+ (NameFormatter *)nameFormatterWithString:(NSString *)aString type:(int)aType {
	NameFormatter *ret;
	
	if(ret = [[[NameFormatter alloc] init] autorelease]) {
		[ret setString:aString];
		[ret setType:aType];
	}
	
	return ret;
}

- (NSString *)string {
	return string;
} 

- (void)setString:(NSString *)aString {
	[string release];
	string = [aString retain];
}

- (int)type {
	return type;
}

- (void)setType:(int)aType {
	type = aType;
}
@end

@implementation BlockingMetadataQuery
- (void) gatheringComplete:(id)notification { 
	if([notification object] != self) {
		NSLogL1(@"in gatheringComplete, got not equals to self");
		return;
	}
	
	[gatheringLock unlock];
	[self enableUpdates];
	
	NSLogL1(@"gatheringComplete");
}

- (void)refreshFinderForItems:(NSArray *)items {
	NSString *firstPath = nil;
	NSEnumerator *propagatedItemsEnumerator = [items objectEnumerator];
	MDNode *propagatedItem;
	
	NSLogL1(@"refreshing items in finder: %@", items);

	while(propagatedItem = [propagatedItemsEnumerator nextObject]) {
		NSString *command, *path = [@"/Volumes/localhost" stringByAppendingPathComponent:[propagatedItem path]];

		if((firstPath == nil) || ([path hasPrefix:firstPath] == NO)) {			
			firstPath = path;
			command = [NSString stringWithFormat:@"ls \"%@\"", [path stringByDeletingLastPathComponent]]; //stringByDeletingLastPathComponent
			NSLogL1(@"going to run ls: %@", command);
			system([command UTF8String]);
		
			if([[propagatedItem query] type] & Folder) //folder alias POSIX
				command = [NSString stringWithFormat:@"osascript -e 'tell application \"Finder\" to update folder POSIX file \"%@\"'", path];
			else
				command = [NSString stringWithFormat:@"osascript -e 'tell application \"Finder\" to update file POSIX file \"%@\"'", path];

			NSLogL1(@"going to prerun: %@", command);
			system([command UTF8String]);
		}
	} 
}

static void MyCallBack(CFNotificationCenterRef center, void *observer, CFStringRef name, const void *object, CFDictionaryRef userInfo) {
	NSArray *keys = [NSArray arrayWithObjects:@"kMDQueryUpdateAddedItems", @"kMDQueryUpdateRemovedItems", @"kMDQueryUpdateChangedItems", nil];
	NSEnumerator *keyEnumerator = [keys objectEnumerator];
	NSString *key;
	extern MDTree *tree;
	
	while(key = [keyEnumerator nextObject]) {
		NSEnumerator *itemEnumerator = [[userInfo objectForKey:key] objectEnumerator];
		NSMetadataItem *item;
		
		while(item = [itemEnumerator nextObject]) {
			NSDictionary *attributes = [item attributeDictionary];
		
			if([key isEqualToString:@"kMDQueryUpdateAddedItems"]) {
				NSLogL1(@"got kMDQueryUpdateAddedItems for %@", [attributes objectForKey:@"kMDItemPath"]);
				[observer refreshFinderForItems:[tree propagateItem:attributes]];
			}
			else if([key isEqualToString:@"kMDQueryUpdateRemovedItems"]) {
				NSLogL1(@"got kMDQueryUpdateRemovedItems for %@", [attributes objectForKey:@"kMDItemPath"]);
				[observer refreshFinderForItems:[tree removeItem:attributes]];
			}
			else if([key isEqualToString:@"kMDQueryUpdateChangedItems"]) {
				NSArray *removedNodes, *propagatedNodes, *intersection;
				NSLogL1(@"got kMDQueryUpdateChangedItems for %@", [item attributeDictionary]);
				removedNodes = [tree removeItem:attributes];	
				propagatedNodes = [tree propagateItem:attributes];
				intersection = [removedNodes filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"NOT (path IN %@.path)", propagatedNodes]];
				intersection = [intersection arrayByAddingObjectsFromArray:[propagatedNodes filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"NOT (path IN %@.path)", removedNodes]]];
				NSLog(@"removedNodes: %@, propedNodes: %@, intersection: %@", removedNodes, propagatedNodes, intersection);
				[observer refreshFinderForItems:intersection];
			}
		}
	}

	NSLogL1(@"at the end of MyCallBack, %@", userInfo);
}

- (void) thread:(id)sender {
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

	double resolution = 0.3;
	BOOL isRunning, ret;
	NSNotificationCenter *center = [NSNotificationCenter defaultCenter];

//	NSLogL1(@"in startQuery, _queryString: %@", [self _queryString]);

	ret = [super startQuery];

	if(!ret) {
		NSLogL1(@"in startQuery, could not startQuery");
		return ret;
	}

	[center addObserver:self selector:@selector(gatheringComplete:) name:NSMetadataQueryDidFinishGatheringNotification object:nil];

	CFNotificationCenterRef nc = CFNotificationCenterGetLocalCenter();
	CFNotificationCenterAddObserver(nc, self, &MyCallBack, kMDQueryDidUpdateNotification, NULL, CFNotificationSuspensionBehaviorDeliverImmediately);

	[[NSRunLoop currentRunLoop] run];
	[pool release];
	
	NSLogL1(@"exiting thread...");
}

- (BOOL) startQuery {
	gatheringLock = [[NSLock alloc] init];
	
	//obtain a lock
	[gatheringLock lock];
	
	NSLogL1(@"starting thread");
	[NSThread detachNewThreadSelector:@selector(thread:) toTarget:self withObject:nil];
	
	//wait for the lock to be released
	[gatheringLock lock];
	
	return YES;
}
@end


@interface NSArray (compare)
- (NSComparisonResult)compare:(id)object;
@end


@implementation NSArray (compare)
- (NSComparisonResult) compare:(id)object {
	NSLogL1(@"comparing...");
	
	if([object isKindOfClass:[NSArray class]]) {
		NSLogL1(@"comparing two arrays, first objects: %@ to %@", [self objectAtIndex:0], [object objectAtIndex:0]);
		return [[self objectAtIndex:0] compare:[object objectAtIndex:0]];
	}
	
	return NSOrderedDescending;
}
@end


/*********************
 * QueryNode
 *********************/

@implementation QueryNode
- (id)init {
	if(![super init]) return nil;
	children = nil;
	return self;
}

- (void) encodeWithCoder: (NSCoder *)coder
{
	[coder encodeObject:[self predicate] forKey:@"predicate"];
	[coder encodeObject:[self nameFormat] forKey:@"nameFormat"];
	[coder encodeInt:(int)[self type] forKey:@"type"];
	[coder encodeObject:[self children] forKey:@"children"];
}

- (id) initWithCoder: (NSCoder *)coder
{
	if (self = [self init])
	{
		[self setNameFormat:[coder decodeObjectForKey:@"nameFormat"]];
		[self setPredicate:[coder decodeObjectForKey:@"predicate"]];
		[self setType:[coder decodeIntForKey:@"type"]];
		children = [[coder decodeObjectForKey:@"children"] retain];
	}

	return self;
}

/*name = [NameFormatter nameFormatterWithString:@"kMDItemAlbum" type:NameFormatterKey];
[queryAlbum2 setNameFormat:[NSArray arrayWithObject:name]];
[queryAlbum2 setPredicate:[NSPredicate predicateWithFormat:@"kMDItemContentTypeTree CONTAINS 'public.audio' AND kMDItemAlbum != nil"]];	
[queryAlbum2 setType:DynamicFolder];
[queryAlbum2 setChild:queryItemName];
*/

- (id)initWithXML:(NSXMLElement *)node {
	if(self = [self init]) {
		if([[node name] isEqualToString:@"node"] || [[node name] isEqualToString:@"leaf"]) {
			NSArray *allMetadataAttibutes = MDSchemaCopyAllAttributes();
			NSXMLNode *attribute;
		
			if(attribute = [node attributeForName:@"predicate"]) {
				[self setPredicate:[NSPredicate predicateWithFormat:[attribute stringValue]]];
			}
		
			if(attribute = [node attributeForName:@"name"]) {
				NSMutableArray *nameFormatters = [[[NSMutableArray alloc] init] autorelease];
				NSCharacterSet *characterSet = [NSCharacterSet characterSetWithCharactersInString:@"@"];
				NSScanner *scanner = [NSScanner scannerWithString:[attribute stringValue]];
			
				[scanner setCharactersToBeSkipped:nil];
			
				while ([scanner isAtEnd] == NO) {
					NameFormatter *nameFormatter;
					NSString *staticText = nil, *key = nil;
					unsigned scanLocation;
					BOOL hasFalseKey = NO;

					if(![scanner scanUpToCharactersFromSet:characterSet intoString:&staticText]) {
						scanLocation = [scanner scanLocation];
						if([scanner scanString:@"@" intoString:NULL] && 
							[scanner scanUpToCharactersFromSet:characterSet intoString:&key]) {
					
							if([allMetadataAttibutes containsObject:key] && [scanner scanString:@"@" intoString:NULL]) {
								nameFormatter = [NameFormatter nameFormatterWithString:key type:NameFormatterKey];
								[nameFormatters addObject:nameFormatter];
							}
							else hasFalseKey = YES;
						}
						else if(scanLocation != [scanner scanLocation]) hasFalseKey = YES;
				
						if(hasFalseKey == YES) {
							NSString *falseKey;
					
							if(falseKey = [[scanner string] substringWithRange:NSMakeRange(scanLocation, ([scanner scanLocation] - scanLocation))]) {
								if(staticText) staticText = [staticText stringByAppendingString:falseKey];
								else staticText = falseKey;
							}	
						}
					}
				
					if(staticText) {
						nameFormatter = [NameFormatter nameFormatterWithString:staticText type:NameFormatterText];
						[nameFormatters addObject:nameFormatter];
					}
				}
			
				[self setNameFormat:nameFormatters];

				if([[node name] isEqualToString:@"node"]) {
					if([[nameFormatters filteredArrayUsingPredicate:[NSPredicate predicateWithFormat:@"type == %d", NameFormatterKey]] count] == 0)
						[self setType:StaticFolder];
					else
						[self setType:DynamicFolder];
				}
				else [self setType:DynamicFile];
			}
			else {
				NSLogL1(@"element %@ (%@) must have a name", [node XPath], [node XMLString]);
				exit(0);
			}
		}
		else if([[node name] isEqualToString:@"tree"]) {
			[self setType:StaticFolder];
			[self setNameFormat:[NSArray arrayWithObject:[NameFormatter nameFormatterWithString:@"/" type:NameFormatterText]]];
		}
		else {
			NSLogL1(@"unknown element %@ (%@)", [node XPath], [node XMLString]);
			exit(0);
		}
		
		int count = [node childCount], i;
		for(i=0; i < count; i++) {
			QueryNode *queryChild = [[QueryNode alloc] initWithXML:[node childAtIndex:i]];
			[self setChild:queryChild];
		}
	}
	
	return self;
}

- (void) writeGraphVizRepresentationTo:(NSString *)path {
	[[self graphVizRepresentation] writeToFile:path atomically:NO encoding:NSUTF8StringEncoding error:nil];
}

- (NSString *)graphVizRepresentation {
	return [NSString stringWithFormat:@"digraph G {\nnode [shape = \"Mrecord\"];\n%@}", [self graphVizRepresentation:@"/"]];
}

- (NSString *)graphVizRepresentation:(NSString *)path {
	NSMutableString *ret = [[NSMutableString alloc] init];
	NSEnumerator *enumerator = [[self children] objectEnumerator];
	QueryNode *child;
	NSString *predicateString;
	int i=0;
	
	if(!(predicateString = [[[self predicate] description] stringWithStringReplaced:@"\"" with:@"'"])) predicateString = @"No Predicate";
	[ret appendFormat:@"\"%@\" [label = \"{%@|%@|%@}\"];\n", path, predicateString, [[[self nameFormat] rawDescription] stringWithStringReplaced:@"\"" with:@"'"], [self typeString]];
	
	while(child = [enumerator nextObject]) {
		NSString *childPath = [path stringByAppendingPathComponent:[NSString stringWithFormat:@"%d", i]];
		[ret appendString:[child graphVizRepresentation:childPath]];
		[ret appendFormat:@"\"%@\" -> \"%@\";\n", path, childPath];
		i++;
	}
	
	return ret;
}

- (NSPredicate *)predicate {
	return predicate;
}

- (void)setPredicate:(NSPredicate *)aPredicate {
	[predicate release];
	predicate = [aPredicate retain];
}

- (NSString *)staticName {
	NSEnumerator *enumerator = [nameFormat objectEnumerator];
	NameFormatter *formatter;
	NSMutableString *name = [[NSMutableString alloc] init];
	
	while(formatter = [enumerator nextObject]) {
		if([formatter type] == NameFormatterText)
			[name appendString:[formatter string]];
	}
	
	return name;
}

- (NSArray *)nameFormat {
	return nameFormat;
}

- (void)setNameFormat:(NSArray *)aNameFormat {
	[nameFormat release];
	nameFormat = [aNameFormat retain];
}

- (int)type {
	return type;
}

- (NSString *)typeString {
	switch([self type]) {
		case DynamicFolder:
			return @"Dynamic Folder";
		case DynamicFile:
			return @"Dynamic File";
		case StaticFolder:
			return @"Static Folder";
	}
}

- (void)setType:(int) aType {
	type = aType;
}

- (void)setChild:(QueryNode *)child {
	if(!children) children = [[NSMutableArray alloc] init];
	[children addObject:child];
}

- (QueryNode *)childAtIndex:(unsigned)index {
	if(children && ([children count] > index)) return [children objectAtIndex:index];
	else return nil;
}

- (void)removeChildAtIndex:(unsigned)index {
	if(children && ([children count] > index)) [children removeObjectAtIndex:index];
}

- (NSArray *)children {
	return children;
}

- (int)childCount {
	return (children != nil) ? [children count] : 0;
}

- (BOOL)isDynamic {
	return ([self type] & Dynamic);
}
@end

/*********************
 * MDNode
 *********************/

@implementation MDNode
- (id)init {
	if(self = [super init]) {
		children = nil;
		query = nil;
		item = nil;
		name = nil;
		parent = nil;
	}
	
	return self;
}

- (void) encodeWithCoder: (NSCoder *)coder
{
	[coder encodeObject:name forKey:@"name"];
	[coder encodeObject:item forKey:@"item"];
	[coder encodeObject:query forKey:@"query"];
	[coder encodeObject:children forKey:@"children"];
	[coder encodeObject:parent forKey:@"parent"];
	NSLogL1(@"encode children: %@", children);
}

- (id) initWithCoder: (NSCoder *)coder
{
	if (self = [self init])
	{
		[self setName:[coder decodeObjectForKey:@"name"]];
		[self setItem:[coder decodeObjectForKey:@"item"]];
		[self setQuery:[coder decodeObjectForKey:@"query"]];
		[self setParent:[coder decodeObjectForKey:@"parent"]];
		children = [[coder decodeObjectForKey:@"children"] retain];
		NSLogL1(@"decode children: %@", children);
	}

	return self;
}

- (NSString *)description {
	return [NSString stringWithFormat:@"MDNode: %@", [self path]];
}

- (NSString *)formattedName {
	NSMutableString *formattedName = [[NSMutableString alloc] init];
	NSEnumerator *enumerator = [[query nameFormat] objectEnumerator];
	NameFormatter *formatter;
	
	while(formatter = [enumerator nextObject]) {
		if([formatter type] == NameFormatterKey) {
			id value = [item objectForKey:[formatter string]];
			
			/* this is a temporary hack. something needs to be come up with to address an array's elements within nameFormat */
			if([value isKindOfClass:[NSArray class]])
				[formattedName appendString:[value objectAtIndex:0]];
			else if([value isKindOfClass:[NSString class]])
				[formattedName appendString:value];
			else if([value isKindOfClass:[NSNumber class]])
				[formattedName appendString:[value description]];
			
			NSLogL1(@"just appended... string as it stands:%@", formattedName);
		}
		else if([formatter type] == NameFormatterText)
			[formattedName appendString:[formatter string]];
		//duplicate handler goes here
	}
	
	return formattedName;
}

/* temporary hack until another predicate variable (other than $DirectoryName) is established to address a directories keys */
- (id) nameObject {
	NSDictionary *dict = (NSDictionary *)MDSchemaCopyMetaAttributesForAttribute((CFStringRef)[[[[self query] nameFormat] objectAtIndex:0] string]);
   int type = [[dict objectForKey:@"kMDAttributeType"] intValue];

	NSLogL1(@"here: %d %@", type, [[[[self query] nameFormat] objectAtIndex:0] string]);

   if(type == CFStringGetTypeID())
		return [self name];
	else if(type == CFNumberGetTypeID())
		return [NSNumber numberWithInt:[[self name] intValue]];
	else
		return [self name];
}

- (NSString *)name {
	return name;
}

- (void)setName:(NSString *)aName {
	[name release];
	name = [aName retain];
}

- (QueryNode *)query {
	return query;
}

- (void)setQuery:(QueryNode *)aQuery {
	[query release];
	query = [aQuery retain];
}

- (NSDictionary *)item {
	return item;
}

- (void)setItem:(NSDictionary *)aItem {
	[item release];
	item = [aItem retain];
}

- (NSMutableDictionary *)children {
	return children;
}

- (MDNode *)child:(NSString *)aName {
	NSEnumerator *enumerator = [children keyEnumerator];
	NSString *child;
	
	/* big time hack...needs to be fixed by keeping all the keys lowercased */
	while(child = [enumerator nextObject]) {
		if([[child lowercaseString] isEqualToString:[aName lowercaseString]]) {
			return [children objectForKey:child];
		}
	}
	
	return nil;
}

- (int)childCount {
	return (!children) ? 0 : [children count];
}

- (MDNode *)parent {
	return parent;
}

- (void)setParent:(MDNode *)aParent {
	[parent release];
	parent = [aParent retain];
}

- (NSString *)path {
	NSMutableArray *pathComponents = [[NSMutableArray alloc] init];
	MDNode *current = self;
	
	do {
		[pathComponents insertObject:[current name] atIndex:0];
	} while(current = [current parent]);
	
	return [NSString pathWithComponents:pathComponents];
}

- (void)setChild:(NSString *)aName child:(MDNode *)child {
	if(!children) children = [[NSMutableDictionary alloc] init];
	
	[child setParent:self];
	[children setObject:child forKey:aName];
}

/* figures out the name based on the item and query and inserts it as a child */
- (MDNode *)setChild:(NSDictionary *)aItem query:(QueryNode *)aQuery isDuplicate:(BOOL *)isDuplicate {
	MDNode *child = [[MDNode alloc] init], *previousChild;
	[child setQuery:aQuery];
	[child setItem:aItem];
	
	NSMutableString *childName = (NSMutableString *)[child formattedName];
	[childName replaceOccurrencesOfString:@"/" withString:@":" options:nil range:NSMakeRange(0, [childName length])];
	
	[child setName:childName];
	if(!(previousChild = [self child:childName])) {
		[self setChild:childName child:child];
		(*isDuplicate) = NO;
	}
	else {
		(*isDuplicate) = YES;
		return previousChild;
	}
	
	return child;
}

- (void)removeChild:(NSString *)aName {
	[children removeObjectForKey:aName];
}
@end

/*********************
 * MDTree
 *********************/

@implementation MDTree
- (void) encodeWithCoder: (NSCoder *)coder
{
	[coder encodeObject:root forKey:@"root"];
}

- (id) initWithCoder: (NSCoder *)coder
{
	if (self = [self init])
	{
		root = [[coder decodeObjectForKey:@"root"] retain];
	}

	return self;
}

+ (MDTree *)treeWithRoot:(MDNode *)aRoot {
	MDTree *ret;
	
	if(ret = [[[MDTree alloc] init] autorelease]) {
		ret->root = aRoot;
	}
	
	return ret;
}

- (MDNode *)findNode:(NSString *)path {
	NSArray *pathComponents = [path pathComponents];
	NSEnumerator *enumerator = [pathComponents objectEnumerator];
	NSString *pathComponent;
	MDNode *currentNode = root;
	
	/* skip over the first slash */
	[enumerator nextObject];
	
	while((pathComponent = [enumerator nextObject]) && (currentNode = [currentNode child:pathComponent]));
	
	if(!pathComponent) return currentNode;
	else return nil;
}

- (NSArray *)removeItem:(NSDictionary *)item {
	NSLog(@"in removeItem: %@", [root children]);
	
	return [self removeItem:item node:root];
}

- (NSArray *)removeItem:(NSDictionary *)item node:(MDNode *)node {
	NSMutableArray *ret = [[[NSMutableArray alloc] init] autorelease];
	NSString *path = [item objectForKey:@"kMDItemPath"];
	NSEnumerator *enumerator = [[[node children] allValues] objectEnumerator];
	MDNode *child;
	
	NSLogL1(@"in new removeItem: for %@, %@", [node path], [node children]);
	
	while(child = [enumerator nextObject]) {
		NSLogL1(@"testing %@", [child path]);
		
		if([[[child item] objectForKey:@"kMDItemPath"] isEqualToString:path] && ([[child query] type] == DynamicFile)) {
			NSLogL1(@"removing %@, %@, %@", [child path], [child name], [[child item] objectForKey:@"kMDItemPath"]);
			[node removeChild:[child name]];
			[ret addObject:child];
		}
		else
			[ret addObjectsFromArray:[self removeItem:item node:child]];
	}
	
	if(([[node query] type] == DynamicFolder) && ([[node children] count] == 0)) {
		[[node parent] removeChild:[node name]];
		NSLogL1(@"removing all for %@", [node path]);
		[ret insertObject:node atIndex:0];
	}
	
	return ret;
}

- (NSArray *)propagateItem:(NSDictionary *)item {
	[self propagateItem:item node:root predicate:nil];
}

- (NSArray *)propagateItem:(NSDictionary *)item node:(MDNode *)node predicate:(NSPredicate *)predicate {
	NSMutableArray *ret = [[NSMutableArray alloc] init];
	QueryNode *query = [node query];
	int i = 0;
	
	NSLogL1(@"nodeName:%@ queryName:%@ predicate:%@, children: %@, item: %@", [node name], [query staticName], [predicate predicateFormat], [node children], [item objectForKey:@"kMDItemPath"]);
	
	NSEnumerator *enumerator = [[query children] objectEnumerator];
	QueryNode *queryChild;
	
	while(queryChild = [enumerator nextObject]) {
		NSPredicate *childPredicate;
	
		if(!predicate)
			childPredicate = [queryChild predicate];
		else if(![queryChild predicate])
			childPredicate = predicate;
		else {
			NSPredicate *tmpChildPredicate;
		
			NSLog(@"node nameObject: %@", [node nameObject]);
			tmpChildPredicate = [[queryChild predicate] predicateWithSubstitutionVariables:[NSDictionary dictionaryWithObject:[node nameObject] forKey:@"DirectoryName"]];
			NSArray *childPredicateArray = [NSArray arrayWithObjects:predicate, tmpChildPredicate, nil];
			childPredicate = [NSCompoundPredicate andPredicateWithSubpredicates:childPredicateArray];
		}
	
		if(([query childCount] == 1) && (queryChild = [query childAtIndex:0]) && [queryChild isDynamic]) {
			NSLogL1(@"childPredicate:%@", [childPredicate predicateFormat]);

			BOOL result = NO;
		
			@try {
				result = [childPredicate evaluateWithObject:item];
			}
			@catch ( NSException *e ) {
				NSLogL1(@"exception... %@", e);
				goto end;
			}

			if(result) {
				NSArray *propagatedItems;
				MDNode *child;
				BOOL isDuplicate = YES;

				child = [node setChild:item query:queryChild isDuplicate:&isDuplicate];
			
				/* setChild: will never overwrite a currently existing child, if one exists it will be returned and then propagated */
				propagatedItems = [self propagateItem:item node:child predicate:childPredicate];
			
				if(isDuplicate == NO) {		
					NSLogL1(@"adding in dynamic if: %@, %@", [child path], child);
					[ret addObject:child];
				}
			
				if(propagatedItems) [ret addObjectsFromArray:propagatedItems];
			}
		}
		else if([query childCount] > 0) {
			NSLogL1(@"in static");
		
			MDNode *child = [node child:[queryChild staticName]];
			BOOL isDuplicate = YES;
			NSArray *propagatedItems;
		
			NSLogL1(@"queryChild: %@", [queryChild staticName]);
		
			if(!child) {
				child = [node setChild:item query:queryChild isDuplicate:&isDuplicate];
			}
		
			propagatedItems = [self propagateItem:item node:child predicate:childPredicate];

			if(isDuplicate == NO) {
				NSLogL1(@"adding in static if: %@", [child path]);
				[ret addObject:child];
			}
		
			NSLogL1(@"adding to ret array: %@", propagatedItems);
		
			if(propagatedItems) [ret addObjectsFromArray:propagatedItems];
		}
	}
	
	NSLogL1(@"return");
	
end:
	if([ret count] <= 0) {
		[ret release];
		ret = nil;
	}
	else 
		[ret autorelease];
	
	return ret;
}

//- (BOOL)isItemApplicable:(NSMetadataItem *)item predicate:(NSPredicate *)predicate directoryName

@end

/*********************
 * Metaphor
 *********************/

@implementation WebDAV
webdav_property live_properties[] = {
	{"DAV:", "creationdate"},
	{"DAV:", "displayname"},
	{"DAV:", "getcontentlanguage"},
	{"DAV:", "getcontentlength"},
	{"DAV:", "getcontenttype"},
	{"DAV:", "getetag"},
	{"DAV:", "getlastmodified"},
	{"DAV:", "resourcetype"},
	{"DAV:", "lockdiscovery"},
	{"DAV:", "source"},
	{"DAV:", "supportedlock"},
	{NULL, NULL}
};

+ (void)allProperties:(MDNode *)node connection:(connection *)con buffer:(buffer *)b {
	MDNode *child;
	
	if([node childCount] <= 0) {
		[self propertiesFromNode:node connection:con buffer:b];
	}
	else {
		NSEnumerator *enumerator = [[node children] objectEnumerator];
		while(child = [enumerator nextObject]) {
			[self propertiesFromNode:child connection:con buffer:b];
		}
	}
	
	return;
}

+ (void)propertiesFromNode:(MDNode *)node connection:(connection *)con buffer:(buffer *)b {
	buffer_append_string(b, "<D:response>\n");
	buffer_append_string(b, "<D:href>");
	buffer_append_string_buffer(b, con->uri.scheme);
	buffer_append_string(b, "://");
	buffer_append_string_buffer(b, con->uri.authority);
	buffer_append_string_encoded(b, CONST_BUF_LEN(con->uri.path), ENCODING_REL_URI);

	NSString *cleanName = [[node name] stringByEncodingHTTPURL];

	buffer_append_string_len(b, [cleanName UTF8String], [cleanName length]);
	buffer_append_string(b, "</D:href>\n");
	buffer_append_string(b, "<D:propstat>\n");
	buffer_append_string(b, "<D:prop>\n");

	if([[node query] type] & Folder) [self directoryProperties:node connection:con buffer:b];
	else [self fileProperties:node connection:con buffer:b];	

	buffer_append_string(b, "</D:prop>\n");
	buffer_append_string(b, "<D:status>HTTP/1.1 200 OK</D:status>\n");
	buffer_append_string(b, "</D:propstat>\n");
	buffer_append_string(b, "</D:response>\n");
	
	return;
}

+ (void)directoryProperties:(MDNode *)node connection:(connection *)con buffer:(buffer *)b  {
	int i = 0;

	for (i = 0; live_properties[i].prop; i++) {
		char ctime_buf[] = "2005-08-26T07:27:16Z";
		char mtime_buf[] = "Thu, 26 Aug 2005 07:27:16 GMT";

		if (0 == strcmp(live_properties[i].prop, "resourcetype")) {
			buffer_append_string(b,
								 "<D:resourcetype><D:collection/></D:resourcetype>");
			continue;
		} else if (0 == strcmp(live_properties[i].prop, "getcontenttype")) {
			buffer_append_string(b,
								 "<D:getcontenttype>httpd/unix-directory</D:getcontenttype>");
			continue;
		} else if (0 == strcmp(live_properties[i].prop, "creationdate")) {
			buffer_append_string(b, "<D:creationdate ns0:dt=\"dateTime.tz\">");
			buffer_append_string(b, ctime_buf);
			buffer_append_string(b, "</D:creationdate>");
			continue;
		} else if (0 == strcmp(live_properties[i].prop, "getlastmodified")) {
			buffer_append_string(b,
								 "<D:getlastmodified ns0:dt=\"dateTime.rfc1123\">");
			buffer_append_string(b, mtime_buf);
			buffer_append_string(b, "</D:getlastmodified>");
			continue;
		} else if (0 == strcmp(live_properties[i].prop, "getcontentlength")) {
			buffer_append_string(b, "<D:getcontentlength>");
			buffer_append_off_t(b, 64);
			buffer_append_string(b, "</D:getcontentlength>");
			continue;
		} else if (0 == strcmp(live_properties[i].prop, "getcontentlanguage")) {
			buffer_append_string(b, "<D:getcontentlanguage>");
			buffer_append_string(b, "en");
			buffer_append_string(b, "</D:getcontentlanguage>");
		}
	}

	return;
}

+ (void)fileProperties:(MDNode *)node connection:(connection *)con buffer:(buffer *)b {
	NSDictionary *item = [node item];
	char ctime_buf[] = "2005-08-18T07:27:16Z";
	char mtime_buf[] = "Thu, 18 Aug 2005 07:27:16 GMT";
	size_t k;
	int i = 0;
	
	for(i=0; live_properties[i].prop; i++) {
		if (strcmp(live_properties[i].prop, "resourcetype") == 0) {
			//when would it be a directory?
		} 
		else if (strcmp(live_properties[i].prop, "getcontenttype") == 0) {
			for (k = 0; k < con->conf.mimetypes->used; k++) {
				data_string *ds = (data_string *) con->conf.mimetypes->data[k];

				if (ds->key->used == 0)
					continue;

				if([[[item objectForKey:@"kMDItemDisplayName"] pathExtension] isEqualToString:[NSString stringWithCString:(ds->key->ptr+1)]] == YES) {
					buffer_append_string(b, "<D:getcontenttype>");
					buffer_append_string_buffer(b, ds->value);
					buffer_append_string(b, "</D:getcontenttype>");
					break;
				}
			}
		}
		else if (strcmp(live_properties[i].prop, "creationdate") == 0) {
			buffer_append_string(b, "<D:creationdate ns0:dt=\"dateTime.tz\">");
			//strftime(ctime_buf, sizeof (ctime_buf), "%Y-%m-%dT%H:%M:%SZ", gmtime(&(sce->st.st_ctime)));
			NSDate *date = [item objectForKey:@"kMDItemFSCreationDate"];
			char *tmp = (char *)[[date descriptionWithCalendarFormat:@"%Y-%m-%dT%H:%M:%SZ" timeZone:nil locale:nil] UTF8String];
			buffer_append_string(b, tmp);
			buffer_append_string(b, "</D:creationdate>");
		} 
		else if (strcmp(live_properties[i].prop, "getlastmodified") == 0) {
			buffer_append_string(b, "<D:getlastmodified ns0:dt=\"dateTime.rfc1123\">");
			//strftime(mtime_buf, sizeof (mtime_buf), "%a, %d %b %Y %H:%M:%S GMT", gmtime(&(sce->st.st_mtime)));
			NSDate *date = [item objectForKey:@"kMDItemFSContentChangeDate"];
			char *tmp = (char *)[[date descriptionWithCalendarFormat:@"%a, %d %b %Y %H:%M:%S GMT" timeZone:nil locale:nil] UTF8String];
			buffer_append_string(b, tmp);
			buffer_append_string(b, "</D:getlastmodified>");
		} 
		else if (strcmp(live_properties[i].prop, "getcontentlength") == 0) {
			buffer_append_string(b, "<D:getcontentlength>");
			//buffer_append_off_t(b, sce->st.st_size);
			buffer_append_string(b, [[[item objectForKey:@"kMDItemFSSize"] description] UTF8String]);
			buffer_append_string(b, "</D:getcontentlength>");
		} 
		else if (strcmp(live_properties[i].prop, "getcontentlanguage") == 0) {
			buffer_append_string(b, "<D:getcontentlanguage>");
			buffer_append_string(b, "en");
			buffer_append_string(b, "</D:getcontentlanguage>");
		}
	}
}
@end

/*
int tests() {
	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
	BlockingMetadataQuery *query = [[BlockingMetadataQuery alloc] init];
	MDNode *root = [[MDNode alloc] init];
	
	QueryNode *queryRoot = [[QueryNode alloc] init];
	QueryNode *queryArtist = [[QueryNode alloc] init];
	QueryNode *queryAuthors = [[QueryNode alloc] init];
	QueryNode *queryAlbum = [[QueryNode alloc] init];
	QueryNode *queryItemName = [[QueryNode alloc] init];

	NameFormatter *name;

	name = [NameFormatter nameFormatterWithString:@"kMDItemDisplayName" type:NameFormatterKey];
	[queryItemName setNameFormat:[NSArray arrayWithObject:name]];
	[queryItemName setPredicate:[NSPredicate predicateWithFormat:@"kMDItemAlbum == $DirectoryName"]];
	[queryItemName setType:DynamicFile];

	name = [NameFormatter nameFormatterWithString:@"kMDItemAlbum" type:NameFormatterKey];
	[queryAlbum setNameFormat:[NSArray arrayWithObject:name]];
	[queryAlbum setPredicate:[NSPredicate predicateWithFormat:@"kMDItemAuthors[FIRST] == $DirectoryName"]];	
	[queryAlbum setType:DynamicFolder];
	[queryAlbum setChild:queryItemName];
	
	name = [NameFormatter nameFormatterWithString:@"kMDItemAuthors" type:NameFormatterKey];
	[queryAuthors setNameFormat:[NSArray arrayWithObject:name]];
	[queryAuthors setPredicate:[NSPredicate predicateWithFormat:@"kMDItemKind == 'MP3 Audio File'"]];
	[queryAuthors setType:DynamicFolder];
	[queryAuthors setChild:queryAlbum];
	
	name = [NameFormatter nameFormatterWithString:@"Artist" type:NameFormatterText];
	[queryArtist setNameFormat:[NSArray arrayWithObject:name]];
	[queryArtist setType:StaticFolder];
	[queryArtist setChild:queryAuthors];

	name = [NameFormatter nameFormatterWithString:@"/" type:NameFormatterText];
	[queryRoot setNameFormat:[NSArray arrayWithObject:name]];
	[queryRoot setType:StaticFolder];
	[queryRoot setChild:queryArtist];
	
	[root setName:@"/"];
	[root setQuery:queryRoot];

	[query setPredicate:[NSPredicate predicateWithFormat:@"kMDItemKind == 'MP3 Audio File' AND kMDItemAuthors LIKE[c] 'Bob Dylan' AND kMDItemAlbum LIKE[c] '*'"]];
	[query startQuery];
	
	MDTree *tree = [MDTree treeWithRoot:root];
	
	int i=0, count=[query resultCount];
	
	for(i=0; i < count; i++) {
		NSMetadataItem *item = [query resultAtIndex:i];
		NSLogL1(@"Count: %d, Artist: %@, Album: %@, Display Name: %@", i,	
																							[item valueForAttribute:@"kMDItemAuthors"], 
																							[item valueForAttribute:@"kMDItemAlbum"],
																							[item valueForAttribute:@"kMDItemDisplayName"]);	
		[tree propagateItem:item];
		NSLogL1(@"done with prop\n\n");
		NSLogL1(@"%d/%d", i, count);
	}

	[pool release];
	return 0;
}*/
