#import <Cocoa/Cocoa.h>
#import "mod_metaphor_webdav.h"

/* adds a category to NSString for the method removeTrailingSlash, which returns the original string with the trailing slash removed if it exists 
 * if the trailing slash does not exist then the original string is returned unmodified. */
@interface NSString(removeTrailingSlash)
- (NSString *) removeTrailingSlash;
@end

@interface NSString(HTTPURLProcessing)
- (NSString *)stringByEncodingHTTPURL;
- (NSString *)stringByDecodingHTTPURL;
@end

typedef enum { 
	Dynamic = 0x000F,
	Folder = 0x0F00,
	DynamicFolder = 0x0F0F, 
	DynamicFile = 0x00FF, 
	StaticFolder = 0x0F00 
} QueryNodeType;

@interface QueryNode : NSObject <NSCoding> {
	NSPredicate *predicate;
	NSMutableArray *children;
	NSArray *nameFormat;
	short type;
}

- (NSPredicate *)predicate;
- (void)setPredicate:(NSPredicate *)aPredicate;
- (NSArray *)nameFormat;
- (void)setNameFormat:(NSArray *)aNameFormat;
- (int)type;
- (void)setType:(int)aType;
- (void)setChild:(QueryNode *)child;
- (QueryNode *)childAtIndex:(unsigned)index;
- (void)removeChildAtIndex:(unsigned)index;
- (NSArray *)children;
- (int)childCount;
- (BOOL)isDynamic;
@end

@interface MDNode : NSObject <NSCoding> {
	NSString *name;
	NSDictionary *item;
	QueryNode *query;
	NSMutableDictionary *children;
	MDNode *parent;
}

- (NSString *)name;
- (void)setName:(NSString *)aName;
- (QueryNode *)query;
- (void)setQuery:(QueryNode *)aQuery;
- (NSMutableDictionary *)children;
- (NSDictionary *)item;
- (void)setItem:(NSDictionary *)aItem;
- (MDNode *)child:(NSString *)name;
- (int)childCount;
- (void)setChild:(NSString *)name child:(MDNode *)child;
- (MDNode *)setChild:(NSDictionary *)aItem query:(QueryNode *)aQuery;
- (void)removeChild:(NSString *)name;
@end

typedef enum { 
	NameFormatterKey=1, 
	NameFormatterText, 
	NameFormatterDuplicateHandler 
} NameFormatterType;

@interface NameFormatter : NSObject <NSCoding> {
	NSString *string;
	int type;
}

+ (NameFormatter *)nameFormatterWithString:(NSString *)aString type:(int)aType;
- (NSString *)string;
- (void)setString:(NSString *)aString;
- (int)type;
- (void)setType:(int)aType;
@end

@interface MDTree : NSObject <NSCoding> {
	MDNode *root;
}

+ (MDTree *)treeWithRoot:(MDNode *)aRoot;
- (MDNode *)findNode:(NSString *)path;
- (NSArray *)propagateItem:(NSDictionary *)item;
- (NSArray *)propagateItem:(NSDictionary *)item node:(MDNode *)node predicate:(NSPredicate *)predicate;
@end

@interface BlockingMetadataQuery : NSMetadataQuery
{
	NSLock *gatheringLock;
}

- (BOOL) startQuery;
@end

@interface WebDAV : NSObject {
}

+ (void)fileProperties:(MDNode *)node connection:(connection *)con buffer:(buffer *)b;
+ (void)directoryProperties:(MDNode *)node connection:(connection *)con buffer:(buffer *)b;
+ (void)propertiesFromNode:(MDNode *)node connection:(connection *)con buffer:(buffer *)b;
+ (void)allProperties:(MDNode *)node connection:(connection *)con buffer:(buffer *)b;
@end
