#include <sys/types.h>
#include <sys/param.h>
#include <sys/ucred.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/mman.h>

#include "stream.h"

#import "mod_metaphor.h"
#import "mod_metaphor_webdav.h"
#import "mod_metaphor_debug.h"

NSAutoreleasePool *pool;
MDTree *tree;
NSString *dbPath;

/* init the plugin data */
INIT_FUNC(mod_metaphor_init) {
	int i = 0;
	plugin_data *p;
	
	p = calloc(1, sizeof (*p));

	p->tmp_buf = buffer_init();

	p->uri.scheme = buffer_init();
	p->uri.path_raw = buffer_init();
	p->uri.path = buffer_init();
	p->uri.authority = buffer_init();

	p->physical.path = buffer_init();
	p->physical.rel_path = buffer_init();
	p->physical.doc_root = buffer_init();
	p->physical.basedir = buffer_init();

 	pool = [[NSAutoreleasePool alloc] init];

	BlockingMetadataQuery *query = [[BlockingMetadataQuery alloc] init];
	MDNode *root = [[MDNode alloc] init];
	QueryNode *queryRoot = [[QueryNode alloc] init];

	NSString *configPath = [@"~/Library/Application Support/Metaphor/Tree Configuration.xml" stringByExpandingTildeInPath];
/*	if(![[NSFileManager defaultManager] fileExistsAtPath:configPath]) {
		[[NSFileManager defaultManager] createDirectoryAtPath:[configPath stringByDeletingLastPathComponent] attributes:nil];
		[[NSFileManager defaultManager] copyPath:@"../etc/Tree Configuration.xml" toPath:configPath handler:nil];		
	}*/
	
	if([[NSFileManager defaultManager] fileExistsAtPath:configPath]) {	
		NSError *error;
		NSXMLDocument *xmlDoc = [[NSXMLDocument alloc] initWithContentsOfURL:[NSURL fileURLWithPath:configPath] options:nil error:&error];
		
		if(!xmlDoc) {
			NSLogL1(@"Parsing error while opening tree configuration: %@", error);
			exit(0);
		}
		
		queryRoot = [[QueryNode alloc] initWithXML:[xmlDoc rootElement]];
		
		//[queryRoot writeGraphVizRepresentationTo:@"/tmp/treeConfiguration.dot"];
		//exit(1);
		//NSLogL1(@"read from tree config: %@", [NSDictionary dictionaryWithContentsOfFile:configPath]);
		//NSLogL1(@"query root: %@, %d, %d", [queryRoot nameFormat], StaticFolder, [queryRoot type]);
	}
	else {
		NSLog(@"'Tree Configuration.xml' not found at %@", configPath);
		exit(0);
	}
	
	[root setName:@"/"];
	[root setQuery:queryRoot];

	//[query setPredicate:[NSPredicate predicateWithFormat:@"(kMDItemContentTypeTree == 'public.audio') && (kMDItemAuthors LIKE[c] 'Bob Dylan') && ((kMDItemAlbum LIKE[c] 'Blood on the Tracks') || (kMDItemAlbum LIKE[c] 'Blonde on Blonde'))"]];
	[query setPredicate:[NSPredicate predicateWithFormat:@"(kMDItemContentTypeTree == 'public.audio') && (!(kMDItemContentType LIKE[c] 'com.apple.mail.emlx')) && (!(kMDItemContentType LIKE[c] 'public.vcard'))"]];

	NSLogL1(@"starting query");
	[query startQuery];
	
	dbPath = [[@"~/Library/Application Support/Metaphor/metadata.db" stringByExpandingTildeInPath] retain];
	if([[NSFileManager defaultManager] fileExistsAtPath:dbPath]) {
		tree = [[NSKeyedUnarchiver unarchiveObjectWithFile:dbPath] retain];
		if(tree) {
			for(i=0; i < [query resultCount]; i++) {
				[query resultAtIndex:i];
			}
			NSLogL1(@"done getting items for cache");
			
			return p;
		}
	}
	
	tree = [MDTree treeWithRoot:root];
	
	int count=[query resultCount];
	
	for(i=0; i < count; i++) {
		//NSLog(@"propagate: %@", [[[query resultAtIndex:i] attributeDictionary] objectForKey:@"kMDItemPath"]);
		[tree propagateItem:[[query resultAtIndex:i] attributeDictionary]];
	}

	return p;
}

/* detroy the plugin data */
FREE_FUNC(mod_metaphor_free) {
	plugin_data *p = p_d;
	
	[NSKeyedArchiver archiveRootObject:tree toFile:dbPath];

	UNUSED(srv);

	if (!p)
		return HANDLER_GO_ON;

	if (p->config_storage) {
		size_t i;
		for (i = 0; i < srv->config_context->used; i++) {
			plugin_config *s = p->config_storage[i];

			if (!s)
				continue;

			free(s);
		}
		free(p->config_storage);
	}
	buffer_free(p->uri.scheme);
	buffer_free(p->uri.path_raw);
	buffer_free(p->uri.path);
	buffer_free(p->uri.authority);

	buffer_free(p->physical.path);
	buffer_free(p->physical.rel_path);
	buffer_free(p->physical.doc_root);
	buffer_free(p->physical.basedir);

	buffer_free(p->tmp_buf);

	free(p);

	return HANDLER_GO_ON;
}

/* handle plugin config and check values */

SETDEFAULTS_FUNC(mod_metaphor_set_defaults) {
	plugin_data *p = p_d;
	size_t i = 0;

	/* TODO: learn about the config SCOPE */
	config_values_t cv[] = {
		{"metaphor.activate", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION},
		{"metaphor.log-xml", NULL, T_CONFIG_BOOLEAN, T_CONFIG_SCOPE_CONNECTION},
		{NULL, NULL, T_CONFIG_UNSET, T_CONFIG_SCOPE_UNSET}
	};

	if (!p)
		return HANDLER_ERROR;

	p->config_storage =
		calloc(1, srv->config_context->used * sizeof (specific_config *));

	for (i = 0; i < srv->config_context->used; i++) {
		plugin_config *s;

		s = calloc(1, sizeof (plugin_config));

		cv[0].destination = &(s->enabled);
		cv[1].destination = &(s->log_xml);

		p->config_storage[i] = s;

		if (0 != config_insert_values_global(srv,
											 ((data_config *) srv->
											  config_context->data[i])->value,
											 cv)) {
			return HANDLER_ERROR;
		}
	}
	
	return HANDLER_GO_ON;
}

#define PATCH(x) \
	p->conf.x = s->x;
static int
mod_metaphor_patch_connection(server * srv, connection * con, plugin_data * p) {
	size_t i, j;
	plugin_config *s = p->config_storage[0];

	PATCH(enabled);

	/* skip the first, the global context */
	for (i = 1; i < srv->config_context->used; i++) {
		data_config *dc = (data_config *) srv->config_context->data[i];
		s = p->config_storage[i];

		/* condition didn't match */
		if (!config_check_cond(srv, con, dc))
			continue;

		/* merge config */
		for (j = 0; j < dc->value->used; j++) {
			data_unset *du = dc->value->data[j];

			if (buffer_is_equal_string
				(du->key, CONST_STR_LEN("metaphor.activate"))) {
				PATCH(enabled);
			} else if (buffer_is_equal_string(du->key,
											  CONST_STR_LEN
											  ("metaphor.log-xml"))) {
				PATCH(log_xml);
			}
		}
	}

	return 0;
}

#undef PATCH

URIHANDLER_FUNC(mod_metaphor_uri_handler) {
	plugin_data *p = p_d;

	UNUSED(srv);

	if (con->uri.path->used == 0)
		return HANDLER_GO_ON;

	mod_metaphor_patch_connection(srv, con, p);
	if (!p->conf.enabled)
		return HANDLER_GO_ON;

	switch (con->request.http_method) {
	case HTTP_METHOD_OPTIONS:
		/*
		 * we fake a little bit but it makes MS W2k happy and it
		 * let's us mount the volume 
		 */
		 
		response_header_overwrite(srv, con, CONST_STR_LEN("DAV"),
								  CONST_STR_LEN("1"));
		response_header_overwrite(srv, con, CONST_STR_LEN("MS-Author-Via"),
								  CONST_STR_LEN("DAV"));
		response_header_insert(srv, con, CONST_STR_LEN("Allow"),
							   CONST_STR_LEN("PROPFIND"));
		break;

	default:
		break;
	}

	/* not found */
	return HANDLER_GO_ON;
}

URIHANDLER_FUNC(mod_metaphor_subrequest_handler) {
	plugin_data *p = p_d;
	buffer *b;
	data_string *ds;
	int depth = -1;
	webdav_properties *req_props;

	UNUSED(srv);

	if (!p->conf.enabled)
		return HANDLER_GO_ON;

	/* physical path is setup */
	if (con->physical.path->used == 0)
		return HANDLER_GO_ON;

	/* PROPFIND need them */
	if ((ds = (data_string *)array_get_element(con->request.headers, "Depth")) != NULL)
		depth = strtol(ds->value->ptr, NULL, 10);

	switch (con->request.http_method) {
		case HTTP_METHOD_PROPFIND: {
			NSString *path = [[[NSString stringWithCString:con->uri.path_raw->ptr] removeTrailingSlash] stringByDecodingHTTPURL];
			
			if([[path lastPathComponent] hasPrefix:@"."] == YES) {
				con->http_status = 404;
				return HANDLER_FINISHED;
			}
			
			MDNode *node = [tree findNode:path];
			
			if (!node) {
				con->http_status = 404;
				return HANDLER_FINISHED;
			}

			/* they want to know the properties of the directory */
			req_props = NULL;
			con->http_status = 207;
			response_header_overwrite(srv, con, CONST_STR_LEN("Content-Type"), CONST_STR_LEN("text/xml; charset=\"utf-8\""));
			b = chunkqueue_get_append_buffer(con->write_queue);
			buffer_copy_string(b, "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n");
			buffer_append_string(b, "<D:multistatus xmlns:D=\"DAV:\" xmlns:ns0=\"urn:uuid:c2f41010-65b3-11d1-a29f-00aa00c14882/\">\n");
								
			[WebDAV allProperties:node connection:con buffer:b];

			/* does this ever happen? */
			if (req_props) {
				size_t i;
				
				for (i = 0; i < req_props->used; i++) {
					free(req_props->ptr[i]->ns);
					free(req_props->ptr[i]->prop);
					free(req_props->ptr[i]);
				}
				free(req_props->ptr);
				free(req_props);
			}
			buffer_append_string(b, "</D:multistatus>\n");

			if (p->conf.log_xml) {
				log_error_write(srv, __FILE__, __LINE__, "sb", "XML-response-body:", b);
			}
			con->file_finished = 1;

			return HANDLER_FINISHED;
			break;
		}
	
		case HTTP_METHOD_MKCOL:
		case HTTP_METHOD_DELETE:
		case HTTP_METHOD_PUT:
		case HTTP_METHOD_MOVE:
		case HTTP_METHOD_COPY:
		case HTTP_METHOD_PROPPATCH:
			con->http_status = 403;
			return HANDLER_FINISHED;
			
		default:
			break;
	}

	/* not found */
	return HANDLER_GO_ON;
}

URIHANDLER_FUNC(mod_metaphor_docroot) {
	if (con->request.http_method != HTTP_METHOD_GET)
		return HANDLER_GO_ON;
	
	NSString *path = [[NSString stringWithCString:con->uri.path_raw->ptr] stringByDecodingHTTPURL];
	MDNode *node = [tree findNode:path];
	NSString *realPath = nil;
	
	if(node && ([[node query] type] == DynamicFile)) realPath = [[node item] objectForKey:@"kMDItemPath"];

	if (realPath) {
		buffer *root = buffer_init_string("/");
		buffer *rel_path = buffer_init_string([realPath UTF8String]);

		buffer_copy_string_buffer(con->physical.doc_root, root);
		buffer_copy_string_buffer(con->physical.rel_path, rel_path);

		buffer_free(rel_path);
		buffer_free(root);
		return HANDLER_GO_ON;
	}

	con->http_status = 404;
	return HANDLER_FINISHED;
}

/* this function is called at dlopen() time and inits the callbacks */

int
mod_metaphor_plugin_init(plugin * p) {
	p->version = LIGHTTPD_VERSION_ID;
	p->name = buffer_init_string("metaphor");

	p->init = mod_metaphor_init;
	p->handle_uri_clean = mod_metaphor_uri_handler;
	p->handle_physical = mod_metaphor_subrequest_handler;
	p->handle_docroot = mod_metaphor_docroot;
	p->set_defaults = mod_metaphor_set_defaults;
	p->cleanup = mod_metaphor_free;

	p->data = NULL;

	return 0;
}
